-- --------------------------------------------------------------------
-- **                            Functions                           **
-- --------------------------------------------------------------------

local levels = {
	CHANGE = 4,
	ABSORB = 5,
	NECRO = 5,
	OVERTIME = 1,
}

local function cap(minValue, value, maxValue)
    return max(minValue, min(maxValue, value))
end

local function EnhancedBars_GetMode(mode, delta)
	if mode == "CHANGE" then
        if delta > 0 then
			return "INCREASE"
		else
			return "DECREASE"
		end
	elseif mode == "OVERTIME" then
        if delta > 0 then
			return "HOT"
		else
			return "DOT"
		end
	else
		return mode
	end
end

function EnhancedBars_CreateBar(parent, mode, isCompact)
    local level = levels[mode]
    assert(level)
	
    local bar = CreateFrame("Frame", nil, parent)
    bar.mode = mode
	bar.style = isCompact and 1 or 0
    bar.guid = nil
    bar.pause = 0
    bar.active = false
    bar.currentValue = 0
    bar.targetValue = 0
    bar.maxValue = 0
    bar.speed = 0
	bar.texture = bar:CreateTexture(nil, "OVERLAY")
	bar.texture:SetTexture(isCompact and "Interface\\RaidFrame\\Raid-Bar-Hp-Fill" or "Interface\\TargetingFrame\\UI-StatusBar")
	bar.texture:SetAllPoints(bar)
	bar:SetScript("OnUpdate", EnhancedBars_OnUpdate)
    bar:SetFrameStrata(parent:GetFrameStrata())
    bar:SetFrameLevel(parent:GetFrameLevel() + level)
	bar:Hide()

    return bar
end

local function EnhancedBars_Snap(bar)
	bar.active = false
	bar.pause = 0
    bar.currentValue = bar.targetValue
end

function EnhancedBars_VerifyGUID(bar, guid)
    if bar.guid ~= guid then
        bar.guid = guid
		EnhancedBars_Snap(bar)
    end
end

local function EnhancedBars_FinalizeBar(bar, shownValue)
	local parent = bar:GetParent()
    local delta = shownValue - bar.currentValue
	local width = parent:GetWidth()
	local height = 1
	local reversed = false
	local anchor = "BOTTOMLEFT"
    local ratio = 0

	if bar.mode == "ABSORB" then
		height = 0.5
elseif bar.mode == "NECRO" then
		height = 0.5
		anchor = "BOTTOMRIGHT"
elseif delta >= 0 then
        ratio = bar.currentValue / bar.maxValue
		reversed = true
  else
        ratio = shownValue / bar.maxValue
    end
	
	if reversed then
		bar.texture:SetTexCoord(1, 0, 0, 1)
	else
		bar.texture:SetTexCoord(0, 1, 0, 1)
	end

	bar:SetWidth(min(1, abs(delta) / max(1, bar.maxValue)) * width)
    bar:SetHeight(parent:GetHeight() * height)
    bar:SetScale(parent:GetScale())
    bar:SetPoint(anchor, parent, anchor, ratio * width, 0)
    bar:Show()
end

function EnhancedBars_ModifyChangeBar(bar, value, maxValue)
    assert(bar.mode == "CHANGE")
    maxValue = max(1, maxValue)
    value = cap(0, value, maxValue)

    bar.currentValue = min(bar.currentValue, maxValue) -- Never overflow the max
    bar.targetValue = value
    bar.maxValue = maxValue
	
    local delta = bar.targetValue - bar.currentValue
    if delta ~= 0 then
        local mode = EnhancedBars_GetMode(bar.mode, delta)
        local minSpeed, maxSpeed, adjustTime, pause, color = EnhancedBars_GetLayout(bar.style, mode)
        local speed = cap(minSpeed, math.abs(delta) / (adjustTime * maxValue), maxSpeed)

        if not bar.active then
            bar.pause = pause
            bar.speed = speed
            bar.active = true
			
			-- Filter out trivial/HoT healing
			local pct = delta / maxValue
			if pct > 0 and pct < 0.05 then
				EnhancedBars_Snap(bar)
			end
      else
            bar.speed = max(speed, bar.speed)
        end

        bar.texture:SetVertexColor(color.r, color.g, color.b)
        EnhancedBars_FinalizeBar(bar, bar.targetValue)
  else
        bar.active = false
        bar:Hide()
    end
end

function EnhancedBars_ModifyPredictedBar(bar, predictedValue, value, maxValue)
    assert(bar.mode == "ABSORB" or bar.mode == "NECRO" or bar.mode == "OVERTIME")
    maxValue = max(1, maxValue)
    value = cap(0, value, maxValue)
    if not (predictedValue > value and bar.mode == "ABSORB") then
        predictedValue = cap(0, predictedValue, maxValue)
    end

    bar.currentValue = value
	bar.targetValue = value
    bar.maxValue = maxValue

    local delta = predictedValue - value
    if delta ~= 0 then
        local mode = EnhancedBars_GetMode(bar.mode, delta)
        local minSpeed, maxSpeed, adjustTime, pause, color = EnhancedBars_GetLayout(bar.style, mode)

		bar.active = true
        bar.texture:SetVertexColor(color.r, color.g, color.b)
        EnhancedBars_FinalizeBar(bar, predictedValue)
  else
		bar.active = false
        bar:Hide()
    end
end

function EnhancedBars_OnUpdate(bar, elapsed)
	if bar.mode == "CHANGE" then
		local delta = bar.targetValue - bar.currentValue
		if delta ~= 0 then
			if bar.pause > 0 then
				bar.pause = max(0, bar.pause - elapsed)
		  else
				local change = bar.speed * bar.maxValue * elapsed
				if delta > 0 then
					bar.currentValue = min(bar.targetValue, bar.currentValue + change)
			  else
					bar.currentValue = max(bar.targetValue, bar.currentValue - change)
				end
			end
			EnhancedBars_FinalizeBar(bar, bar.targetValue)
	  else
			bar.active = false
			bar:Hide()
		end
	end
end