-- ========================================================
--                           Data
-- ========================================================

local treatAsEffect = {
	-- [Hit spell] = Effect spell
	[186439] = 187464, -- Shadow Mend
}

local channelPseudoEffect = {
	-- [Hit spell] = Channel spell name
	[212105] = 212084, -- Fel Devastation
	[198030] = 198013, -- Eye Beam
	[258926] = 258925, -- Fel Barrage
	[117418] = 113656, -- Fists of Fury
	[107270] = 101546, -- Spinning Crane Kick
	
	-- Needs to be tested before adding:
	--[50622] = 227847, -- Bladestorm (handle off-hand carefully -- also different spells for Arms/Fury)
	-- ...more?
}

local channelLookup = { }

do
	for k, v in pairs(channelPseudoEffect) do
		channelLookup[GetSpellInfo(v)] = v
	end
end

local absorbs = { }
local channeling = { }
local cache = { }
local effects = { }

local scaling = { }
local measures = { }

local NEXT_POLL = 0
local MEASURE_SCALING = false

local POLLED_UNITS = {
	"player",
	"pet",
	"pettarget",
	"target",
	"targettarget",
	"focus",
	"focustarget",
	"mouseover",
	"mouseovertarget",
}

do
	for i=1, 5 do
		tinsert(POLLED_UNITS, "boss"..i)
		tinsert(POLLED_UNITS, "arena"..i)
	end
	for i=1, 20 do
		tinsert(POLLED_UNITS, "nameplate"..i)
	end
end

local function ScalingRatio(guid)
	local s = scaling[guid]
	if s then
		return s[4]
	end
	return 1
end

-- ========================================================
--                           API
-- ========================================================

OverTimePredictor_Version = 101

-- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
-- Use powerType < 0 to get health prediction
-- Returns total, dot, hot if unit is valid
-- Returns 0, 0, 0 if unit is invalid
-- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function OverTimePredictor_GetBalance(powerType, unit)
	if powerType < 0 then
		local guid = UnitGUID(unit)
		if guid then
			local u = effects[guid]
			if u then
				local hot, dot = 0, 0
				for i=1, #u do
					local e = u[i]
					if e[10] < 0 then
						hot = hot - e[10]
					else
						dot = dot + e[10]
					end
				end
				local s = ScalingRatio(guid)
				hot = hot * s
				dot = dot * s
				return hot - dot, hot, dot
			end
		end
	end
	return 0, 0, 0
end

-- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
-- Returns total, absorb, necro if unit is valid
-- Returns 0, 0, 0 if unit is invalid
-- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function OverTimePredictor_GetAbsorb(unit)
	local guid = UnitGUID(unit)
	if guid then
		local absorb = UnitGetTotalAbsorbs(unit) + (absorbs[guid] or 0)
		local necro = UnitGetTotalHealAbsorbs(unit)

		return absorb - necro, absorb, necro
	end
	return 0, 0, 0
end

-- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
-- Returns PVP scaling coefficient (default 1)
-- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function OverTimePredictor_GetScaling(unit)
	local guid = UnitGUID(unit)
	if guid then
		return ScalingRatio(guid)
	end
	return nil -- Invalid unit
end

-- ========================================================
--                         Utility
-- ========================================================

local tables = { }

local function NewTable()
	if #tables > 0 then
		return tremove(tables, #tables)
	else
		return { }
	end
end

local function RecTable(t)
	wipe(t)
	tinsert(tables, t)
end

-- ========================================================
--                         Scaling
-- ========================================================

local function RegisterScaling(guid, combatLogAmount, observedHealthChange)
	if combatLogAmount * observedHealthChange > 0 then -- Consistency check (non-zero terms + same sign)
		local s = scaling[guid]
		if not s then
			s = {0, 0, 0, 1}
			scaling[guid] = s
		end
		s[1] = s[1] + 1
		s[2] = s[2] + abs(observedHealthChange)
		s[3] = s[3] + abs(combatLogAmount)
		s[4] = s[2] / s[3]
		--print("sample", guid, "combatLog", combatLogAmount, "healthChange", observedHealthChange, "i", s[1], "A", s[2], "B", s[3], "S", s[4])
	end
end

local function ClearMeasures(guid, u)
	for i=1, #u do
		RecTable(u[i])
	end
	RecTable(u)
	measures[guid] = nil
end

local function ClearScaling()
	for guid, s in pairs(scaling) do
		RecTable(s)
		scaling[guid] = nil
	end
	for guid, u in pairs(measures) do
		ClearMeasures(guid, u)
	end
end

local function TestMeasures(guid, u)
	if #u >= 4 then
		for i=2, #u-2 do
			-- Look for a damage event
			local e = u[i]
			if #e == 2 and e[2] ~= 0 then
				-- Look at HP measure before and after this damage event
				local hp1 = u[i-1] -- start HP
				local hp2 = u[i+1] -- end HP
				local after = u[i+2] -- event following end HP
				
				-- Unit must stay valid in both measurements
				-- End HP must be between 2 and max HP
				-- Max HP must not change between both measurements
				if (#hp1 == 4 and hp1[4]) and (#hp2 == 4 and hp2[2] > 1 and hp2[2] < hp2[3] and hp2[4]) and hp1[3] == hp2[3] then
					if hp2[1] - e[1] < 0.025 and after[1] - e[1] > 0.1 then
						RegisterScaling(guid, e[2], hp2[2] - hp1[2])
						ClearMeasures(guid, u)
						return
					end
				end
			end
		end
	end
end

local function AddMeasure(u, m)
	tinsert(u, m)
	if #u > 10 then
		RecTable(tremove(u, 1))
	end
end

local function MeasureUnit(unit)
	if UnitIsPlayer(unit) and not UnitIsUnit(unit, "player") then
		local guid = UnitGUID(unit)
		if guid then
			local u = measures[guid]
			if not u then
				u = NewTable()
				measures[guid] = u
			end
			
			local hpLast = u.l
			local hp = UnitHealth(unit)
			local hpMax = UnitHealthMax(unit)
			local valid = UnitAffectingCombat(unit) and not UnitIsDeadOrGhost(unit)
			
			if (not hpLast) or (hpLast[2] ~= hp) or (hpLast[3] ~= hpMax) or (hpLast[4] ~= valid) then
				local m = NewTable()
				m[1] = GetTime()
				m[2] = hp
				m[3] = hpMax
				m[4] = valid
				u.l = m -- Fast access to last HP measure
				AddMeasure(u, m)
				TestMeasures(guid, u)
			end
		end
	end
end

local function MeasureEvent(guid, amount, unreliable)
	if MEASURE_SCALING then
		local u = measures[guid] -- Only register events for measured units (i.e: players)
		if u then
			local m = NewTable()
			m[1] = GetTime()
			m[2] = unreliable and 0 or -amount -- Negative sign for damage, positive for healing
			AddMeasure(u, m)
		end
	end
end

local function CheckScalingToggle()
	-- Don't bother with PVP scaling bullshit outside of PVP mode
	local flag = UnitIsPVP("player") or UnitIsPVPFreeForAll("player")
	
	if MEASURE_SCALING ~= flag then
		MEASURE_SCALING = flag
		ClearScaling()
	end
end

-- ========================================================
--                         Functions
-- ========================================================

local function ReadCache(i, sourceGUID, destGUID, spellId)
	local root = i..sourceGUID..spellId
	return cache[root..destGUID] or cache[root] or 0
end

local function WriteCache(i, sourceGUID, destGUID, spellId, value)
	if value ~= 0 then
		local root = i..sourceGUID..spellId
		cache[root] = value
		cache[root..destGUID] = value
	end
end

local function UpdateTotal(e)
	if e[5] == 0 or e[7] == 0 or e[7] > 60 then
		e[10] = 0 -- Don't compute effects with undefined or huge duration
	else
		e[10] = e[6] / e[5] * max(0, e[8] - GetTime())
	end
end

local function UpdateEffect(e, timestamp, amount)
	-- Update tick interval
	local interval = timestamp - max(e[3], e[4])
	if interval > 0.01 then -- Some spells deal their first tick on effect application
		if e[5] == 0 then
			e[5] = interval
		else
			e[5] = interval * 0.3 + e[5] * 0.7
		end
	end
	e[4] = timestamp
	e[9] = true -- Request duration update
	
	-- Update tick damage
	if amount ~= 0 then
		if e[6] == 0 then
			e[6] = amount
		else
			e[6] = amount * 0.3 + e[6] * 0.7
		end
	end
	
	-- Update total damage
	UpdateTotal(e)
end

local function HandleAddEffect(timestamp, sourceGUID, destGUID, spellId)
	local u = effects[destGUID]
	if not u then
		u = NewTable()
		effects[destGUID] = u
	end
	local interval = ReadCache(1, sourceGUID, destGUID, spellId)
	local amount = ReadCache(2, sourceGUID, destGUID, spellId)
	local now = GetTime()
	local e = NewTable()
	e[1] = sourceGUID 	-- Effect source GUID
	e[2] = spellId 		-- Effect spell ID
	e[3] = timestamp 	-- Effect timestamp
	e[4] = 0 			-- Last tick timestamp
	e[5] = interval     -- Interval (smoothed with exponential filter)
	e[6] = amount		-- Damage per tick (smoothed with exponential filter)
	e[7] = 0 			-- Effect duration
	e[8] = 0 			-- Effect expiration time
	e[9] = true         -- Scan required flag
	e[10] = 0 			-- Estimated total damage remaining (negative for healing)
	e[11] = now + 10    -- Prune timestamp
	tinsert(u, e)
	if spellId == 122470 and sourceGUID == destGUID then -- Touch of Karma
		absorbs[destGUID] = ReadCache(1, destGUID, "", "maxhp") / 2
	end
	--print("HandleAddEffect", timestamp, sourceGUID, destGUID, spellId)
end

local function HandleRefreshEffect(timestamp, sourceGUID, destGUID, spellId)
	local u = effects[destGUID]
	if u then
		for i=1, #u do
			local e = u[i]
			if e[1] == sourceGUID and e[2] == spellId then
				e[9] = true -- Request duration update
				return true
			end
		end
	end
	--print("HandleRefreshEffect", timestamp, sourceGUID, destGUID, spellId)
	return false
end

local function HandleRemoveEffect(timestamp, sourceGUID, destGUID, spellId)
	local u = effects[destGUID]
	if u then
		for i=1, #u do
			local e = u[i]
			if e[1] == sourceGUID and e[2] == spellId then
				-- Cache for better and faster estimation on next effect
				WriteCache(1, sourceGUID, destGUID, spellId, e[5])
				WriteCache(2, sourceGUID, destGUID, spellId, e[6])
				RecTable(tremove(u, i))
				break
			end
		end
		if spellId == 122470 and sourceGUID == destGUID then -- Touch of Karma
			absorbs[destGUID] = nil
		end
	end
	--print("HandleRemoveEffect", timestamp, sourceGUID, destGUID, spellId)
end

local function HandleTick(timestamp, sourceGUID, destGUID, spellId, amount)
	local u = effects[destGUID]
	if u then
		local best = nil
		for i=1, #u do
			local e = u[i]
			if e[1] == sourceGUID and e[2] == spellId then
				if (not best) or max(e[3], e[4]) < max(best[3], best[4]) then
					best = e
				end
			end
		end
		if best then
			UpdateEffect(best, timestamp, amount)
		end
	end
	--print("HandleTick", timestamp, sourceGUID, destGUID, spellId, amount)
end

local function HandleHit(timestamp, sourceGUID, destGUID, spellId, amount)
	local effectSpellId = treatAsEffect[spellId]
	if effectSpellId then
		HandleTick(timestamp, sourceGUID, destGUID, effectSpellId, amount)
	end
	local channelSpellId = channelPseudoEffect[spellId]
	if channelSpellId then
		if not HandleRefreshEffect(timestamp, sourceGUID, destGUID, channelSpellId) then
			HandleAddEffect(timestamp, sourceGUID, destGUID, channelSpellId)
		end
		HandleTick(timestamp, sourceGUID, destGUID, channelSpellId, amount)
	end
	--print("HandleHit", timestamp, sourceGUID, destGUID, spellId, amount)
end

local function HandleAbsorb(destGUID, absorbed, sourcePlayer)
	if absorbed then
		local value = absorbs[destGUID]
		if value then
			if sourcePlayer then
				absorbed = absorbed * ScalingRatio(destGUID)
			end
			absorbs[destGUID] = max(0, value - absorbed)
		end
		--print("HandleAbsorb", destGUID, absorbed)
	end
end

local function Prune(now)
	for guid, u in pairs(effects) do
		for i=#u, 1, -1 do
			local e = u[i]
			if now > e[11] then
				-- Don't cache if an effect is removed through prune mechanism
				RecTable(tremove(u, i))
			end
		end
		if #u == 0 then
			RecTable(u)
			effects[guid] = nil
		end
	end
end

local function Clear()
	wipe(absorbs)
	wipe(channeling)
	wipe(cache)
	Prune(math.huge)
	ClearScaling()
end

-- ========================================================
--                         Handlers
-- ========================================================

local function ScanUnit(unit)
	local guid = UnitGUID(unit)
	if guid then
		-- Scan effects
		local u = effects[guid]
		if u then
			local m = NewTable()
			-- Build aura map (special care is taken to handle duplicate effects)
			for i=1, 100 do
				local _, _, _, _, duration, expirationTime, _, _, _, spellId = UnitAura(unit, i, "HELPFUL HARMFUL")
				if not spellId then break end
				
				local s = m[spellId]
				if not s then
					s = NewTable()
					m[spellId] = s
				end
				local o = NewTable()
				o[1] = duration or 0
				o[2] = expirationTime or 0
				tinsert(s, o)
			end
			-- Resolve effects to their respective aura
			for i=1, #u do
				local e = u[i]
				local s = m[e[2]]
				if s and #s >= 1 then
					local o = tremove(s, 1)
					if e[9] then
						e[9] = false
						e[7] = o[1]
						e[8] = o[2]
						e[11] = max(e[11], e[8] + 10)
						if e[10] == 0 then
							UpdateTotal(e) -- Initialize total damage if possible
						end
					end
					RecTable(o)
				end
			end
			-- Cleanup
			for spellId, s in pairs(m) do
				for i=1, #s do
					RecTable(s[i])
				end
				RecTable(s)
				m[spellId] = nil
			end
			RecTable(m)
		end

		-- Scan active channel
		local name, _, _, startTimeMS, endTimeMS = UnitChannelInfo(unit)
		if name then
			local channelSpellId = channelLookup[name]
			if channelSpellId then
				local duration = (endTimeMS - startTimeMS) / 1000
				local expirationTime = endTimeMS / 1000

				for _, v in pairs(effects) do
					for i=1, #v do
						local e = v[i]
						if e[1] == guid and e[2] == channelSpellId then
							e[7] = duration
							e[8] = expirationTime
							e[11] = max(e[11], e[8] + 10)
							if e[10] == 0 then
								UpdateTotal(e) -- Initialize total damage if possible
							end
							break
						end
					end
				end
				
				channeling[guid] = channelSpellId
				return
			end
		end
		
		-- Scan stopped channel
		local channelSpellId = channeling[guid]
		if channelSpellId then
			for destGUID in pairs(effects) do
				HandleRemoveEffect(0, guid, destGUID, channelSpellId)
			end

			channeling[guid] = nil
		end
		
		-- Scan max HP
		WriteCache(1, guid, "", "maxhp", UnitHealthMax(unit))
	end
end

local function OnCombatLogEvent(timestamp, event, hideCaster, sourceGUID, sourceName, sourceFlags, sourceRaidFlags, destGUID, destName, destFlags, destRaidFlags, ...)
	if not sourceGUID or not destGUID then return end

	local sourcePlayer = bit.band(sourceFlags, COMBATLOG_OBJECT_TYPE_PLAYER) > 0
	
	local spellId, spellName, spellSchool
	local amount, overkill, school, resisted, blocked, absorbed, critical, glancing, crushing, overhealing
	local missType, isOffHand, amountMissed, _

	if strsub(event, 1, 5) == "SPELL" then
		spellId, spellName, spellSchool = ...

		if event == "SPELL_MISSED" then
			missType, isOffHand, amountMissed = select(4, ...)
			HandleHit(timestamp, sourceGUID, destGUID, spellId, amountMissed or 0)
			if missType == "ABSORB" then 
				HandleAbsorb(destGUID, amountMissed, sourcePlayer)
			end
			
		elseif event == "SPELL_DAMAGE" then
			amount, overkill, school, resisted, blocked, absorbed, critical, glancing, crushing, isOffHand = select(4, ...)
			HandleHit(timestamp, sourceGUID, destGUID, spellId, amount + (resisted or 0) + (blocked or 0) + (absorbed or 0))
			HandleAbsorb(destGUID, absorbed, sourcePlayer)
			MeasureEvent(destGUID, amount, overkill >= 0 or absorbed or not sourcePlayer)

		elseif event == "SPELL_HEAL" then
			amount, overhealing, absorbed, critical = select(4, ...)
			HandleHit(timestamp, sourceGUID, destGUID, spellId, -amount)
			MeasureEvent(destGUID, -amount, overhealing > 0 or absorbed > 0 or not sourcePlayer)
			
		elseif event == "SPELL_PERIODIC_MISSED" then
			missType, isOffHand, amountMissed = select(4, ...)
			HandleTick(timestamp, sourceGUID, destGUID, spellId, amountMissed or 0)
			if missType == "ABSORB" then 
				HandleAbsorb(destGUID, amountMissed, sourcePlayer)
			end

		elseif event == "SPELL_PERIODIC_DAMAGE" then
			amount, overkill, school, resisted, blocked, absorbed, critical, glancing, crushing, isOffHand = select(4, ...)
			HandleTick(timestamp, sourceGUID, destGUID, spellId, amount + (resisted or 0) + (blocked or 0) + (absorbed or 0))
			HandleAbsorb(destGUID, absorbed, sourcePlayer)
			MeasureEvent(destGUID, amount, overkill >= 0 or absorbed or not sourcePlayer)
			
		elseif event == "SPELL_PERIODIC_HEAL" then
			amount, overhealing, absorbed, critical = select(4, ...)
			HandleTick(timestamp, sourceGUID, destGUID, spellId, -amount)
			MeasureEvent(destGUID, -amount, overhealing > 0 or absorbed > 0 or not sourcePlayer)

		elseif event == "SPELL_AURA_APPLIED" then
			HandleAddEffect(timestamp, sourceGUID, destGUID, spellId)

		elseif event == "SPELL_AURA_REFRESH" then
			HandleRefreshEffect(timestamp, sourceGUID, destGUID, spellId)

		elseif event == "SPELL_AURA_REMOVED" then
			HandleRemoveEffect(timestamp, sourceGUID, destGUID, spellId)
		end

	elseif event == "SWING_MISSED" then 
		missType, isOffHand, amountMissed = ...
		if missType == "ABSORB" then 
			HandleAbsorb(destGUID, amountMissed, sourcePlayer)
		end

	elseif event == "SWING_DAMAGE" then 
		amount, overkill, school, resisted, blocked, absorbed, critical, glancing, crushing, isOffHand = ...
		HandleAbsorb(destGUID, absorbed, sourcePlayer)
		MeasureEvent(destGUID, amount, overkill >= 0 or absorbed or not sourcePlayer)

	elseif event == "RANGE_MISSED" then 
		missType, isOffHand, amountMissed = select(4, ...)
		if missType == "ABSORB" then 
			HandleAbsorb(destGUID, amountMissed, sourcePlayer)
		end

	elseif event == "RANGE_DAMAGE" then 
		amount, overkill, school, resisted, blocked, absorbed, critical, glancing, crushing, isOffHand = select(4, ...)
		HandleAbsorb(destGUID, absorbed, sourcePlayer)
		MeasureEvent(destGUID, amount, overkill >= 0 or absorbed or not sourcePlayer)
			
	elseif event == "DAMAGE_SHIELD" then
		spellId, spellName, spellSchool, amount, overkill, school, resisted, blocked, absorbed, critical, glancing, crushing = ...
		HandleAbsorb(destGUID, absorbed, sourcePlayer)
		MeasureEvent(destGUID, amount, overkill >= 0 or absorbed or not sourcePlayer)
			
	elseif event == "DAMAGE_SPLIT" then
		spellId, spellName, spellSchool, amount, overkill, school, resisted, blocked, absorbed, critical, glancing, crushing = ...
		HandleAbsorb(destGUID, absorbed, sourcePlayer)
		MeasureEvent(destGUID, amount, overkill >= 0 or absorbed or not sourcePlayer)
			
	elseif event == "ENVIRONMENTAL_DAMAGE" then
		_, amount, overkill, school, resisted, blocked, absorbed, critical, glancing, crushing = ...
		HandleAbsorb(destGUID, absorbed, sourcePlayer)
		MeasureEvent(destGUID, amount, overkill >= 0 or absorbed or not sourcePlayer)
	end
end

function OverTimePredictor_OnLoad(self)
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	self:RegisterEvent("UNIT_HEALTH_FREQUENT")
	self:RegisterEvent("UNIT_AURA")
	self:RegisterEvent("UNIT_TARGET")
	self:RegisterEvent("PLAYER_TARGET_CHANGED")
	self:RegisterEvent("PLAYER_FOCUS_CHANGED")
	self:RegisterEvent("PLAYER_ENTERING_WORLD")
	self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_START")
	self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_STOP")
	self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_UPDATE")
	self:RegisterEvent("UPDATE_MOUSEOVER_UNIT")
end

function OverTimePredictor_OnEvent(self, event, ...)
	if event == "COMBAT_LOG_EVENT_UNFILTERED" then
		OnCombatLogEvent(CombatLogGetCurrentEventInfo())

	elseif event == "UNIT_HEALTH_FREQUENT" then
		if MEASURE_SCALING then
			local unit = ...
			MeasureUnit(unit)
		end
		
	elseif event == "UNIT_AURA" then
		local unit = ...
		ScanUnit(unit)

	elseif event == "UNIT_TARGET" then
		local unit = ...
		ScanUnit(unit.."target")
		
	elseif event == "UNIT_SPELLCAST_CHANNEL_START" or event == "UNIT_SPELLCAST_CHANNEL_STOP" or event == "UNIT_SPELLCAST_CHANNEL_UPDATE" then
		local unit = ...
		ScanUnit(unit)
		
	elseif event == "PLAYER_TARGET_CHANGED" then
		ScanUnit("target")
		ScanUnit("targettarget")

	elseif event == "PLAYER_FOCUS_CHANGED" then
		ScanUnit("focus")
		ScanUnit("focustarget")

	elseif event == "UPDATE_MOUSEOVER_UNIT" then
		ScanUnit("mouseover")
		ScanUnit("mouseovertarget")

	elseif event == "PLAYER_ENTERING_WORLD" then
		Clear()
	end
end

function OverTimePredictor_OnUpdate(self, elapsed)
	local now = GetTime()

	if now > NEXT_POLL then
		NEXT_POLL = now + 0.25
		for i=1, #POLLED_UNITS do
			ScanUnit(POLLED_UNITS[i])
		end
		Prune(now)
		CheckScalingToggle()
		
	elseif MEASURE_SCALING then
		for i=1, #POLLED_UNITS do
			MeasureUnit(POLLED_UNITS[i])
		end	
	end
end